
/*
** (c) 2004 Herbert Poetzl
**
** V0.01	simple helper
** V0.02	added mon and step
** V0.03	mask delta
**
*/

#define VERSION "V0.03"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdint.h>


#include "syscall.h"
#include "mhelper.h"


static	char	err_flag = 0;

static	char	opt_read = 0;
static	char	opt_write = 0;
static	char	opt_vers = 0;
static	char	opt_mon = 0;

static	char *	cmd_name = NULL;

static	int 	num_addr = 0;
static	int 	num_size = 4;
static	int 	num_count = 0;
static	int 	num_step = 4;

static	int 	cmd_read = CMD_READ_L;
static	int 	cmd_write = CMD_WRITE_L;

static	int 	first = 1;

static	uint32_t    *masks;
static	uint32_t    *value;
static	uint32_t    *triple;
static	unsigned    triple_cur = 0;
static	unsigned    triple_max = 64;

int 	match(uint32_t addr, uint32_t from, uint32_t to)
{
    	uint32_t *pos = triple;
    	int j;
	
	for (j=0; j<triple_cur; j++) {
    	    pos = triple + j*4;
	    
	    if ((pos[0]==addr) 
	    	&& (pos[1]==from) 
		&& (pos[2]==to))
	    	break;
    	}
	if (j<triple_cur)
	    return ++pos[3];
	else {
	    if (triple_cur >= triple_max) {
	    	triple_max *= 2;
		triple = realloc(triple, 
		    triple_max*4*sizeof(uint32_t));
	    }
	    pos = triple + triple_cur*4;
	       
	    pos[0]=addr;
	    pos[1]=from;
	    pos[2]=to;
	    pos[3]=0;
	    triple_cur++;
	}
	return 0;
}


void 	do_clear(int level)
{
    	triple_cur = 0;
	if (level) {
	    memset(value, 0, 
	    	num_count*sizeof(uint32_t));
	    memset(masks, 0, 
	    	num_count*sizeof(uint32_t)*4);
	    first = 1;
    	}
}

void 	do_mask_map(int delta)
{
    	uint32_t *pos, *mask, addr;
    	int i, o, index, val;
	char buf[64];
	static char symb[] = { 
	    '0','-','.','.','0','*','.','o',
	    '1','+','1','.','.','.','.','i',
	     };
	
	for (i=0; i<triple_cur; i++) {
	    pos = triple + 4*i;
	    index = (pos[0] - num_addr)/num_step;
	    mask = masks + 4*index;
	    
	    mask[0] |= pos[1] ^ pos[2];
	    mask[1] |= (pos[1] & pos[2]);
	    mask[2] |= ~(pos[1] | pos[2]);
    	}
    	for (index=0; index<num_count; index++) {
	    mask = masks + 4*index;
	    addr = num_addr + index*num_step;

    	    o = snprintf(buf, sizeof(buf), "0x%08x :", addr);

	    for (i=0; i<8*num_size; i++) {
	    	uint32_t bit = (1 << (8*num_size - i - 1));
		
    	    	if ((i%4)==0)
		    buf[o++] = ' ';
	    	val = ((mask[0] & bit)?1:0) |
		    ((mask[1] & bit)?2:0) |
		    ((mask[2] & bit)?4:0) |
		    ((value[index] & bit)?8:0);
		buf[o++] = symb[val];
	    }
	    buf[o++] = 0;
	    printf("%s\n", buf);
	}
	printf("\n");
}


void 	handle(int ch)
{
    	switch(ch) {
	case 3:     exit(0);		/* CTRL-C */
	case 13:    printf("\n");	break;
	case 'm':   do_mask_map(0);	break;
	case 'M':   do_mask_map(1);	break;
	case 'c':   do_clear(0);	break;
	case 'r':   do_clear(1);	break;
	default:    printf("kbd: %d\n", ch); 
	}
}


#define OPTIONS "+ha:mn:rs:w124V"

int	main(int argc, char *argv[])
{
        extern int optind;
        extern char *optarg;
        int c, errflg = 0;
	int r;


        cmd_name = argv[0];
        while ((c = getopt(argc, argv, OPTIONS)) != EOF) {
            switch (c) {
            case 'h':
                fprintf(stderr,
                    "This is %s " VERSION "\n"
                    "options are:\n"
                    "-h        print this help message\n"
                    "-a <addr> start memory address\n"
                    "-m        monitor memory location\n"
                    "-n <num>  memory size (count)\n"
                    "-r        read memory location\n"
                    "-s <num>  address step size\n"
                    "-w        write memory location\n"
                    "-1        value is one byte in size\n"
                    "-2        value is two bytes in size\n"
                    "-V        verify interface version\n"
                    "--        end of options\n"
                    ,cmd_name);
                exit(0);
                break;
            case 'a':
                num_addr = strtol(optarg, NULL, 0);
                break;
            case 'm':
                opt_mon = 1;
                break;
            case 'n':
                num_count = strtol(optarg, NULL, 0);
                break;
            case 'r':
                opt_read = 1;
                break;
            case 's':
                num_step = strtol(optarg, NULL, 0);
                break;
            case 'w':
                opt_write = 1;
                break;
            case '1':
                num_size = num_step = 1;
		cmd_read = CMD_READ_B;
    	    	cmd_write = CMD_WRITE_B;
                break;
            case '2':
                num_size = num_step = 1;
		cmd_read = CMD_READ_W;
    	    	cmd_write = CMD_WRITE_W;
                break;
            case 'V':
                opt_vers = 1;
                break;
            case '?':
            default:
                errflg++;
                break;
            }
        }
        if (errflg) {
            fprintf(stderr, 
                "Usage: %s -[" OPTIONS "] <val> ...\n"
                "%s -h for help.\n",
                cmd_name, cmd_name);
            exit(2);
        }

	if (opt_vers) {
	    r = mhelper(0, 0, 0);
	    if (r<0)
	    	perror("mhelper");
    	    else 
	    	printf("version: %04x:%04x\n", 
		    (r>>16)&0xFFFF, r&0xFFFF);
	}

	if (opt_mon) {
	    struct termios input;
	    int i,j;
	    char buf[1];
	    
	    value = calloc(num_count, sizeof(uint32_t));
	    masks = calloc(num_count, sizeof(uint32_t)*4);
	    triple = calloc(triple_max, sizeof(uint32_t)*4);
	    
	    fcntl(0, F_SETFL, O_NONBLOCK);
	    tcgetattr(0, &input);
	    cfmakeraw(&input);
	    input.c_oflag |= OPOST;
	    tcsetattr(0, 0, &input);
		    
	    while (1) {
	    	for (i=0; i<num_count; i++) {
		    uint32_t addr = num_addr + num_step*i;
		    
		    r = mhelper(cmd_read, addr, 0);
		    if (first)
		    	value[i] = r;
		    if (value[i] != r) {
			int count = match(addr, value[i], r);
			
			if (!count) {
		    	    printf("0x%08x : %0*x -> %0*x  bit",
			    	addr, num_size*2, value[i], 
				num_size*2, r);
    	    	    	    for (j=0; j<num_size*8; j++) {
			    	uint32_t bit = (1 << j);
				
				if ((value[i] ^ r) & bit)
				    printf(" %d", j);
			    }
			    printf("\n");
			}
			value[i] = r;
		    }
		}
		first = 0;
		usleep(100);
		r = read(0, buf, 1);
		if (r==1)
		    handle(buf[0]);
	    }
	}
	
	
	while ((num_count > 0) || (optind < argc)) {
	    uint32_t value = 0;

    	    if (optind < argc)
            	value = strtol(argv[optind], NULL, 0);

	    if (opt_write) {
		r = mhelper(cmd_write, num_addr, value);
		if (r<0)
		    perror("write");
	    } 
	    if (opt_read) {
		r = mhelper(cmd_read|OPT_VERB, num_addr, 0);
		if (r<0)
		    perror("read");
	    }
	    
	    num_addr += num_step;
	    num_count--;
	    
	    if (num_count > 0) {
	    	if (optind < argc-1) 
		    optind++;
	    } else
	    	optind++;    	
	}
	
	exit((err_flag)?1:0);
}

